@@ -15,6 +15,7 @@ import (
 	"sync"
 	"syscall"
 	"time"
+	"unicode"
 
 	"github.com/gcla/gowid"
 	"github.com/gcla/gowid/gwutil"
@@ -29,8 +30,8 @@ import (
 	"github.com/gcla/gowid/widgets/menu"
 	"github.com/gcla/gowid/widgets/styled"
 	"github.com/gcla/gowid/widgets/text"
-	"github.com/gcla/termshark"
-	"github.com/gcla/termshark/widgets/appkeys"
+	"github.com/gcla/termshark/v2"
+	"github.com/gcla/termshark/v2/widgets/appkeys"
 	"github.com/gdamore/tcell"
 )
 
@@ -46,24 +47,27 @@ type filtStruct struct {
 }
 
 type Widget struct {
-	wrapped           gowid.IWidget
-	opts              Options
-	ed                *edit.Widget     // what the user types into - wrapped by validity styling
-	dropDown          *menu.Widget     // the menu of possible completions
-	dropDownSite      *menu.SiteWidget // where in this widget structure the drop down is rendered
-	validitySite      *holder.Widget   // the widget swaps out the contents of this placeholder on validity changes
-	valid             gowid.IWidget    // what to display when the filter value is valid
-	invalid           gowid.IWidget    // what to display when the filter value is invalid
-	intermediate      gowid.IWidget    // what to display when the filter value's validity is being determined
-	edCtx             context.Context
-	edCancelFn        context.CancelFunc
-	fields            termshark.IPrefixCompleter // provides completions, given a prefix
-	completionsList   *list.Widget               // the filter widget replaces the list walker when new completions are generated
-	completions       []string                   // the current set of completions, used when rendering
-	runthisfilterchan chan *filtStruct
-	filterchangedchan chan *filtStruct
-	quitchan          chan struct{}
-	readytorunchan    chan struct{}
+	wrapped              gowid.IWidget
+	opts                 Options
+	ed                   *edit.Widget     // what the user types into - wrapped by validity styling
+	dropDown             *menu.Widget     // the menu of possible completions
+	dropDownSite         *menu.SiteWidget // where in this widget structure the drop down is rendered
+	validitySite         *holder.Widget   // the widget swaps out the contents of this placeholder on validity changes
+	valid                gowid.IWidget    // what to display when the filter value is valid
+	invalid              gowid.IWidget    // what to display when the filter value is invalid
+	intermediate         gowid.IWidget    // what to display when the filter value's validity is being determined
+	edCtx                context.Context
+	edCancelFn           context.CancelFunc
+	edCtxLock            sync.Mutex
+	fields               termshark.IPrefixCompleter // provides completions, given a prefix
+	completionsList      *list.Widget               // the filter widget replaces the list walker when new completions are generated
+	completionsActivator *activatorWidget           // used to disable focus going to drop down
+	completions          []string                   // the current set of completions, used when rendering
+	runthisfilterchan    chan *filtStruct
+	filterchangedchan    chan *filtStruct
+	quitchan             chan struct{}
+	readytorunchan       chan struct{}
+	temporarilyDisabled  *bool // set to true right after submitting a new filter, so the menu disappears
 	*gowid.Callbacks
 	gowid.IsSelectable
 }
@@ -74,6 +78,7 @@ var _ io.Closer = (*Widget)(nil)
 type IntermediateCB struct{}
 type ValidCB struct{}
 type InvalidCB struct{}
+type SubmitCB struct{}
 
 type Options struct {
 	Completer      termshark.IPrefixCompleter
@@ -84,14 +89,17 @@ func New(opt Options) *Widget {
 	ed := edit.New()
 
 	fixed := gowid.RenderFixed{}
-	l2 := list.New(list.NewSimpleListWalker([]gowid.IWidget{}))
+	filterList := list.New(list.NewSimpleListWalker([]gowid.IWidget{}))
+	filterActivator := &activatorWidget{
+		IWidget: filterList,
+	}
 
 	if opt.MaxCompletions == 0 {
 		opt.MaxCompletions = 20
 	}
 
 	menuListBox2 := styled.New(
-		framed.NewUnicode(cellmod.Opaque(l2)),
+		framed.NewUnicode(cellmod.Opaque(filterActivator)),
 		gowid.MakePaletteRef("filter-menu-focus"),
 	)
 
@@ -111,7 +119,10 @@ func New(opt Options) *Widget {
 		YOffset: 1,
 	})
 
-	onelineEd := appkeys.New(ed, filterOutEnter, appkeys.Options{
+	cb := gowid.NewCallbacks()
+
+	temporarilyDisabled := false
+	onelineEd := appkeys.New(ed, handleEnter(cb, &temporarilyDisabled), appkeys.Options{
 		ApplyBefore: true,
 	})
 
@@ -138,23 +149,25 @@ func New(opt Options) *Widget {
 	filterchangedchan := make(chan *filtStruct)
 
 	res := &Widget{
-		wrapped:           wrapped,
-		opts:              opt,
-		ed:                ed,
-		dropDown:          drop,
-		dropDownSite:      site,
-		validitySite:      placeholder,
-		valid:             valid,
-		invalid:           invalid,
-		intermediate:      intermediate,
-		fields:            opt.Completer,
-		completionsList:   l2,
-		completions:       []string{},
-		filterchangedchan: filterchangedchan,
-		runthisfilterchan: runthisfilterchan,
-		quitchan:          quitchan,
-		readytorunchan:    readytorunchan,
-		Callbacks:         gowid.NewCallbacks(),
+		wrapped:              wrapped,
+		opts:                 opt,
+		ed:                   ed,
+		dropDown:             drop,
+		dropDownSite:         site,
+		validitySite:         placeholder,
+		valid:                valid,
+		invalid:              invalid,
+		intermediate:         intermediate,
+		fields:               opt.Completer,
+		completionsList:      filterList,
+		completionsActivator: filterActivator,
+		completions:          []string{},
+		filterchangedchan:    filterchangedchan,
+		runthisfilterchan:    runthisfilterchan,
+		quitchan:             quitchan,
+		readytorunchan:       readytorunchan,
+		temporarilyDisabled:  &temporarilyDisabled,
+		Callbacks:            cb,
 	}
 
 	validcb := &ValidateCB{
@@ -225,7 +238,12 @@ func New(opt Options) *Widget {
 			// Tell other goroutine we are ready for more - each time round the loop. This makes sure
 			// we don't run more than one tshark process - it will get killed if a new filter should take
 			// priority.
-			res.readytorunchan <- struct{}{}
+			select {
+			case res.readytorunchan <- struct{}{}:
+			case <-res.quitchan:
+				break CL
+			}
+
 			select {
 			case <-res.quitchan:
 				break CL
@@ -238,38 +256,12 @@ func New(opt Options) *Widget {
 		}
 	}, Goroutinewg)
 
+	ed.OnCursorPosSet(gowid.MakeWidgetCallback("cb", gowid.WidgetChangedFunction(func(app gowid.IApp, ew gowid.IWidget) {
+		res.UpdateCompletions(app)
+	})))
+
 	ed.OnTextSet(gowid.MakeWidgetCallback("cb", gowid.WidgetChangedFunction(func(app gowid.IApp, ew gowid.IWidget) {
-		// Shortcut - we know that "" is always valid
-		if ed.Text() != "" {
-			res.validitySite.SetSubWidget(res.intermediate, app)
-			gowid.RunWidgetCallbacks(res.Callbacks, IntermediateCB{}, app, res)
-		}
-
-		if res.edCancelFn != nil {
-			res.edCancelFn()
-		}
-		res.edCtx, res.edCancelFn = context.WithCancel(context.Background())
-
-		// don't kick things off right away in case user is typing fast
-		go func(ctx context.Context) {
-			select {
-			case <-ctx.Done():
-				return
-			case <-time.After(time.Millisecond * 200):
-				break
-			}
-
-			res.filterchangedchan <- &filtStruct{ed.Text(), app}
-
-			app.Run(gowid.RunFunction(func(app gowid.IApp) {
-				_, y := app.GetScreen().Size()
-				makeCompletions(res.fields, ed.Text(), y, app, func(completions []string, app gowid.IApp) {
-					app.Run(gowid.RunFunction(func(app gowid.IApp) {
-						res.processCompletions(completions, app)
-					}))
-				})
-			}))
-		}(res.edCtx)
+		res.UpdateCompletions(app)
 	})))
 
 	return res
@@ -345,13 +337,36 @@ func (f *Validator) Validate(filter string) {
 	}
 }
 
-func filterOutEnter(evk *tcell.EventKey, app gowid.IApp) bool {
-	handled := false
-	switch evk.Key() {
-	case tcell.KeyEnter:
-		handled = true
+// if the filter is valid when enter is pressed, submit the SubmitCB callback. Those
+// registered will be able to respond e.g. start handling the valid filter value.
+func handleEnter(cb *gowid.Callbacks, temporarilyDisabled *bool) appkeys.KeyInputFn {
+	return func(evk *tcell.EventKey, app gowid.IApp) bool {
+		handled := false
+		switch evk.Key() {
+		case tcell.KeyEnter:
+
+			var dummy gowid.IWidget
+			gowid.RunWidgetCallbacks(cb, SubmitCB{}, app, dummy)
+			*temporarilyDisabled = true
+
+			handled = true
+		}
+		return handled
 	}
-	return handled
+}
+
+func isValidFilterRune(r rune) bool {
+	res := true
+	switch {
+	case unicode.IsLetter(r):
+	case unicode.IsNumber(r):
+	case r == '-':
+	case r == '_':
+	case r == '.':
+	default:
+		res = false
+	}
+	return res
 }
 
 func newMenuWidgets(ed *edit.Widget, completions []string) []gowid.IWidget {
@@ -370,7 +385,8 @@ func newMenuWidgets(ed *edit.Widget, completions []string) []gowid.IWidget {
 			button.Options{
 				Decoration:         button.BareDecoration,
 				SelectKeysProvided: true,
-				SelectKeys:         []gowid.IKey{gowid.MakeKeyExt(tcell.KeyEnter)},
+				//SelectKeys:         []gowid.IKey{},
+				SelectKeys: []gowid.IKey{gowid.MakeKeyExt(tcell.KeyEnter)},
 			},
 		)
 		clickmeStyled := styled.NewInvertedFocus(clickme, gowid.MakePaletteRef("filter-menu-focus"))
@@ -378,15 +394,26 @@ func newMenuWidgets(ed *edit.Widget, completions []string) []gowid.IWidget {
 			txt := ed.Text()
 			end := ed.CursorPos()
 			start := end
+		Loop1:
 			for {
 				if start == 0 {
 					break
 				}
-				if start < len(txt) && txt[start] == ' ' {
+				start--
+				if !isValidFilterRune(rune(txt[start])) {
 					start++
+					break Loop1
+				}
+			}
+		Loop2:
+			for {
+				if end == len(txt) {
 					break
 				}
-				start--
+				if !isValidFilterRune(rune(txt[end])) {
+					break Loop2
+				}
+				end++
 			}
 			ed.SetText(fmt.Sprintf("%s%s%s", txt[0:start], scopy, txt[end:len(txt)]), app)
 			ed.SetCursorPos(len(txt[0:start])+len(scopy), app)
@@ -424,10 +451,64 @@ func makeCompletions(comp termshark.IPrefixCompleter, txt string, max int, app g
 	comp.Completions(txt, cb)
 }
 
+// Start an asynchronous routine to update the drop-down menu with completion
+// options. Runs on a small delay so it can be cancelled and restarted if the
+// user is typing quickly.
 func (w *Widget) UpdateCompletions(app gowid.IApp) {
-	makeCompletions(w.fields, "", w.opts.MaxCompletions, app, func(completions []string, app gowid.IApp) {
-		w.processCompletions(completions, app)
-	})
+	app.Run(gowid.RunFunction(func(app gowid.IApp) {
+		if w.ed.Text() != "" {
+			w.validitySite.SetSubWidget(w.intermediate, app)
+			gowid.RunWidgetCallbacks(w.Callbacks, IntermediateCB{}, app, w)
+		}
+	}))
+
+	// UpdateCompletions can be called outside of the app goroutine, so we
+	// need to protect the context
+	w.edCtxLock.Lock()
+	defer w.edCtxLock.Unlock()
+
+	if w.edCancelFn != nil {
+		w.edCancelFn()
+	}
+	w.edCtx, w.edCancelFn = context.WithCancel(context.Background())
+
+	// don't kick things off right away in case user is typing fast
+	go func(ctx context.Context) {
+		select {
+		case <-ctx.Done():
+			return
+		case <-time.After(time.Millisecond * 200):
+			break
+		}
+
+		// Send the value to be run by tshark. This will kill any other one in progress.
+		w.filterchangedchan <- &filtStruct{w.ed.Text(), app}
+
+		app.Run(gowid.RunFunction(func(app gowid.IApp) {
+			_, y := app.GetScreen().Size()
+
+			txt := w.ed.Text()
+			end := w.ed.CursorPos()
+			start := end
+		Loop:
+			for {
+				if start == 0 {
+					break
+				}
+				start--
+				if !isValidFilterRune(rune(txt[start])) {
+					start++
+					break Loop
+				}
+			}
+
+			makeCompletions(w.fields, txt[start:end], y, app, func(completions []string, app gowid.IApp) {
+				app.Run(gowid.RunFunction(func(app gowid.IApp) {
+					w.processCompletions(completions, app)
+				}))
+			})
+		}))
+	}(w.edCtx)
 }
 
 func (w *Widget) processCompletions(completions []string, app gowid.IApp) {
@@ -440,6 +521,9 @@ func (w *Widget) processCompletions(completions []string, app gowid.IApp) {
 	w.completions = completions
 	app.Run(gowid.RunFunction(func(app gowid.IApp) {
 		w.completionsList.SetWalker(list.NewSimpleListWalker(menu2Widgets), app)
+		// whenever there's an update, take focus away from drop down. This means enter
+		// can be used to submit a new filter.
+		w.completionsActivator.active = false
 		w.dropDown.SetWidth(gowid.RenderWithUnits{U: max + 2}, app)
 	}))
 }
@@ -451,6 +535,10 @@ func (w *Widget) Close() error {
 	return nil
 }
 
+func (w *Widget) OnSubmit(f gowid.IWidgetChangedCallback) {
+	gowid.AddWidgetCallback(w, SubmitCB{}, f)
+}
+
 func (w *Widget) OnIntermediate(f gowid.IWidgetChangedCallback) {
 	gowid.AddWidgetCallback(w, IntermediateCB{}, f)
 }
@@ -484,7 +572,10 @@ func (w *Widget) RenderSize(size gowid.IRenderSize, focus gowid.Selector, app go
 }
 
 func (w *Widget) Render(size gowid.IRenderSize, focus gowid.Selector, app gowid.IApp) gowid.ICanvas {
-	if focus.Focus && len(w.completions) > 0 {
+	// It can be disabled if e.g. the user's last input caused the filter value to
+	// be submitted. Then the best UX is to not display the drop down until further input
+	// or cursor movement.
+	if focus.Focus && len(w.completions) > 0 && !*w.temporarilyDisabled {
 		w.dropDown.Open(w.dropDownSite, app)
 	} else {
 		w.dropDown.Close(app)
@@ -492,12 +583,61 @@ func (w *Widget) Render(size gowid.IRenderSize, focus gowid.Selector, app gowid.
 	return w.wrapped.Render(size, focus, app)
 }
 
-// Reject tab because I want it to switch views. Not intended to be transferable.
+// Reject tab because I want it to switch views. Not intended to be transferable. Reject down because
+// accepting it triggers the OnCursorSet callback, which re-evaluates the filter value - the user sees
+// it go orange briefly, which is unpleasant.
 func (w *Widget) UserInput(ev interface{}, size gowid.IRenderSize, focus gowid.Selector, app gowid.IApp) bool {
-	if evk, ok := ev.(*tcell.EventKey); ok && evk.Key() == tcell.KeyTAB {
-		return false
+	if evk, ok := ev.(*tcell.EventKey); ok {
+		if evk.Key() == tcell.KeyTAB || evk.Key() == tcell.KeyDown {
+			return false
+		}
 	}
-	return gowid.UserInput(w.wrapped, ev, size, focus, app)
+	*w.temporarilyDisabled = false // any input should start the appearance of the drop down again
+	return w.wrapped.UserInput(ev, size, focus, app)
+}
+
+//======================================================================
+
+// activatorWidget is intended to wrap a ListBox, and will suppress focus to the listbox by
+// default, which has the effect of not highlighting any listbox items. The intended effect
+// is for the cursor to be "above" the first item. When the user hits down, then focus
+// is passed through, so the top item is highlighted. If the key pressed is up, and the
+// listbox doesn't handle it, that must mean it's at the top of its range, so the effect is
+// start suppressing focus again.
+type activatorWidget struct {
+	gowid.IWidget
+	active bool
+}
+
+func (w *activatorWidget) UserInput(ev interface{}, size gowid.IRenderSize, focus gowid.Selector, app gowid.IApp) bool {
+	if ev, ok := ev.(*tcell.EventKey); ok && !w.active {
+		if ev.Key() == tcell.KeyDown {
+			w.active = true
+			return true
+		} else {
+			return false
+		}
+	}
+	res := w.IWidget.UserInput(ev, size, focus, app)
+	if !res {
+		if ev, ok := ev.(*tcell.EventKey); ok && w.active {
+			if ev.Key() == tcell.KeyUp {
+				w.active = false
+				return true
+			} else {
+				return false
+			}
+		}
+	}
+	return res
+}
+
+func (w *activatorWidget) Render(size gowid.IRenderSize, focus gowid.Selector, app gowid.IApp) gowid.ICanvas {
+	newf := focus
+	if !w.active {
+		newf = gowid.NotSelected
+	}
+	return w.IWidget.Render(size, newf, app)
 }
 
 //======================================================================
